#include <stdlib.h>
#include <stdbool.h>
#include <genmc.h>
#include "queue.h"

#ifndef MAX_THREADS
# define MAX_THREADS 32
#endif

typedef struct node {
	unsigned int value;
	_Atomic(struct node *) next;
} node_t;

__VERIFIER_hp_t *get_free_hp();

static node_t *new_node()
{
	return malloc(sizeof(node_t));
}

static void reclaim(node_t *node)
{
	__VERIFIER_hp_retire(node);
}

void init_queue(queue_t *q, int num_threads)
{
	node_t *dummy = new_node();
	atomic_init(&dummy->next, NULL);
	atomic_init(&q->head, dummy);
	atomic_init(&q->tail, dummy);
}

void enqueue(queue_t *q, unsigned int val)
{
	node_t *tail, *next;

	node_t *node = new_node();
	node->value = val;
	node->next = NULL;

	__VERIFIER_hp_t *hp = get_free_hp();
	while (true) {
		tail = __VERIFIER_hp_protect(hp, &q->tail);
		next = atomic_load_explicit(&tail->next, memory_order_acquire);
		if (tail != atomic_load_explicit(&q->tail, memory_order_acquire))
			continue;

		if (next == NULL) {
#ifdef ENQUEUE_WRITE_BUG
			atomic_store_explicit(&tail->next, node, memory_order_release);
			break;
#elif ENQUEUE_XCHG_BUG
			atomic_exchange_explicit(&tail->next, node, memory_order_release);
			break;
#else
			if (atomic_compare_exchange_strong_explicit(&tail->next, &next, node, memory_order_release, memory_order_release))
				break;
#endif
		} else {
			atomic_compare_exchange_strong_explicit(&q->tail, &tail, next,
								memory_order_release, memory_order_release);
		}
	}
	atomic_compare_exchange_strong_explicit(&q->tail, &tail, node, memory_order_release, memory_order_release);
}

bool dequeue(queue_t *q, unsigned int *retVal)
{
	node_t *head, *tail, *next;
	unsigned ret;

	__VERIFIER_hp_t *hp_head = get_free_hp();
	__VERIFIER_hp_t *hp_next = get_free_hp();
	while (true) {
		head = __VERIFIER_hp_protect(hp_head, &q->head);
		tail = atomic_load_explicit(&q->tail, memory_order_relaxed);
		next = __VERIFIER_hp_protect(hp_next, &head->next);
		if (atomic_load_explicit(&q->head, memory_order_acquire) != head)
			continue;
		if (head == tail) {
			if (next == NULL)
				return false;
			atomic_compare_exchange_strong_explicit(&q->tail, &tail, next,
								memory_order_release, memory_order_release);
		} else {
			ret = next->value;
			if (atomic_compare_exchange_strong_explicit(&q->head, &head, next,
								    memory_order_release, memory_order_release))
				break;
		}
	}
	*retVal = ret;
	reclaim(head);
	return true;
}
