#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdatomic.h>
#include <assert.h>

#include "lib/queue.c"

#ifndef MAX_THREADS
# define MAX_THREADS 32
#endif

#ifndef HP_THREAD_LIMIT
# define HP_THREAD_LIMIT 32
#endif

#ifdef NOISE_ENQ
#define DEFAULT_NOISE_ENQ (NOISE_ENQ)
#else
#define DEFAULT_NOISE_ENQ 1
#endif

#ifdef NOISE_DEQ
#define DEFAULT_NOISE_DEQ (NOISE_DEQ)
#else
#define DEFAULT_NOISE_DEQ 0
#endif

queue_t *queue;
queue_t myqueue;
int num_threads;

int __thread tid;

__VERIFIER_hp_t hps[MAX_THREADS + 1][HP_THREAD_LIMIT];
int __thread __hp_index;

void set_thread_num(int i)
{
	tid = i;
}

int get_thread_num()
{
	return tid;
}

__VERIFIER_hp_t *get_free_hp()
{
	int index = __hp_index++;
	assert(index < HP_THREAD_LIMIT);
	return &hps[tid][index];
}

void *thread_enq(void *param)
{
	int pid = (intptr_t ) param;

	set_thread_num(pid);
	enqueue(queue, 1);
	enqueue(queue, 2);
	return NULL;
}

void *thread_deq(void *param)
{
	int pid = (intptr_t ) param;
	unsigned dequeued[2];

	set_thread_num(pid);

	/* Ensure FIFO */
	__VERIFIER_assume(dequeue(queue, &dequeued[0]));
	__VERIFIER_assume(dequeue(queue, &dequeued[1]));
	assert(!(dequeued[0] == 2 && dequeued[1] == 1));
	return NULL;
}

void *noise_enq(void *param)
{
	int pid = (intptr_t ) param;

	set_thread_num(pid);
	enqueue(queue, 0);
	return NULL;
}

void *noise_deq(void *param)
{
	int pid = (intptr_t ) param;
	unsigned val;

	set_thread_num(pid);
	dequeue(queue, &val);
	return NULL;
}

int main()
{
	pthread_t te, td, noise[DEFAULT_NOISE_ENQ + DEFAULT_NOISE_DEQ + 2];

	queue = &myqueue;
	num_threads = 2 + DEFAULT_NOISE_ENQ + DEFAULT_NOISE_DEQ;

	init_queue(queue, num_threads);

	pthread_create(&te, NULL, thread_enq, (void *) (intptr_t) 1);
	pthread_create(&td, NULL, thread_deq, (void *) (intptr_t) 2);
	int i = 1;
	for (int j = 1; j <= DEFAULT_NOISE_ENQ; j++, i++) {
                if (j == 1)
                        pthread_create(&noise[i], NULL, noise_enq, (void *) (intptr_t) 2 + i);
                else
                        noise[i] = __VERIFIER_spawn_symmetric(noise_enq, (void *) (intptr_t) 2 + i, noise[i-1]);
        }
	for (int j = 1; j <= DEFAULT_NOISE_DEQ; j++, i++) {
                if (j == 1)
                        pthread_create(&noise[i], NULL, noise_deq, (void *) (intptr_t) 2 + i);
                else
                        noise[i] = __VERIFIER_spawn_symmetric(noise_deq, (void *) (intptr_t) 2 + i, noise[i-1]);
	}

	pthread_join(te, NULL);
	pthread_join(td, NULL);
	i = 1;
	for (int j = 1; j <= DEFAULT_NOISE_ENQ; j++, i++) {
                if (j == 1)
                        pthread_join(noise[i], NULL);
                else
                        __VERIFIER_join_symmetric(noise[i]);
	}
	for (int j = 1; j <= DEFAULT_NOISE_DEQ; j++, i++) {
                if (j == 1)
                        pthread_join(noise[i], NULL);
                else
                        __VERIFIER_join_symmetric(noise[i]);
	}
	return 0;
}
